/*
 * testloadself - test local rules
 *
 * Copyright (C) 2010 Casey Schaufler <casey@schaufler-ca.com>
 * Copyright (C) 2012 Intel Corporation
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2.
 *
 * Author:
 *	Casey Schaufler <casey@schaufler-ca.com>
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/xattr.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "smacktools.h"

#define LSIZE 23
#define ASIZE 5
#define TICKY	"TickyTickyTickySpathangWhoo"

int writeself(int loadfd, char *sp, char *op, char *real)
{
	char rule[LSIZE + LSIZE + ASIZE + 3];

	sprintf(rule, "%-23s %-23s %-5s", sp, op, real);

	printf("w\t%s\n", rule);

	return write(loadfd, rule, LSIZE + LSIZE + ASIZE + 2);
}

int writeself2(int load2fd, char *sp, char *op, char *real)
{
	char *rule;
	int rlen;
	
	rlen = strlen(sp) + strlen(op) + strlen(real) + 4;
	if ((rule = malloc(rlen)) == NULL)
		return -1;
	sprintf(rule, "%s %s %s", sp, op, real);

	printf("w\t%s\n", rule);

	rlen = write(load2fd, rule, strlen(rule));
	free(rule);
	return rlen;
}

int offset;
char buffer[1024];

int readself(int loadfd, char *sp, char *op, char *real)
{
	int rc;

	if (offset == 0) {
		rc = read(loadfd, buffer, sizeof(buffer));
		if (rc < 0) {
			perror("reading /smack/load-self");
			return rc;
		}
		if (rc == 0)
			return 0;
		buffer[rc] = '\0';
	}
	if ((rc = sscanf(buffer + offset, "%s %s %s", sp, op, real)) != 3) {
		printf("scanned %d\n", rc);
		rc = -1;
	}
	while (buffer[offset] != '\n' && buffer[offset] != '\0')
		offset++;
	if (buffer[offset] == '\n')
		offset++;
	if (buffer[offset] == '\0')
		offset = 0;
	return rc;
}

int openload(int oldfd)
{
	int loadfd;

	if (oldfd >= 0)
		close(oldfd);

	loadfd = open("/sys/fs/smackfs/load-self", O_RDWR);
	if (loadfd < 0)
		loadfd = open("/smack/load-self", O_RDWR);
	if (loadfd < 0) {
		perror("opening /smack/load-self");
		exit(1);
	}
	return loadfd;
}

int openload2(int oldfd)
{
	int load2fd;

	if (oldfd >= 0)
		close(oldfd);

	load2fd = open("/sys/fs/smackfs/load-self2", O_RDWR);
	if (load2fd < 0)
		load2fd = open("/smack/load-self2", O_RDWR);
	if (load2fd < 0) {
		perror("opening /smack/load-self2");
		exit(1);
	}
	return load2fd;
}

int smack_of_file(char *path, char *smack)
{
	int rc;

	rc = lgetxattr(path, "security.SMACK64", smack, SMACK_LABEL);
	if (rc > 0)
		smack[rc] = '\0';
	return rc;
}

void printlist(int loadfd, char *title, int use2)
{
	char subject[SMACK_LABEL];
	char object[SMACK_LABEL];
	char accesses[ASIZE+1];

	if (use2)
		loadfd = openload2(loadfd);
	else
		loadfd = openload(loadfd);
	printf("%s:\n", title);
	while (readself(loadfd, subject, object, accesses) > 0)
		printf("r\t%s %s %s\n", subject, object, accesses);
}

int
main(int argc, char *argv[])
{
	int i;
	int rc;
	int loadfd;
	int use2 = 0;
	char subject[SMACK_LABEL];
	char object[SMACK_LABEL];
	char accesses[ASIZE+1];
	char self[SMACK_LABEL];
	char it[SMACK_LABEL];
	char *testfile = NULL;

	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-2") == 0 ||
		    strcmp(argv[i], "-l") == 0 ||
		    strcmp(argv[i], "--long") == 0 ||
		    strcmp(argv[i], "--load-self2") == 0)
			use2 = 1;
		if (strcmp(argv[i], "-f") == 0 ||
		    strcmp(argv[i], "--file") == 0)
			testfile = argv[++i];
	}

	loadfd = use2 ? openload2(-1) : openload(-1);

	printf("euid =%d, uid = %d\n", geteuid(), getuid());
	printlist(loadfd, "Initial list", use2);

	loadfd = use2 ? openload2(loadfd) : openload(loadfd);
	printf("Writing\n");

	if (use2) {
		if (writeself2(loadfd, TICKY, "Crackle", "rwxa-") < 0)
			perror("Writing \"" TICKY " Crackle rwxa-\"");
		if (writeself2(loadfd, TICKY, "Crackle", "rwx--") < 0)
			perror("Writing \"" TICKY " Crackle rwx--\"");
		if (writeself2(loadfd, TICKY, "Snap", "rwxat") < 0)
			perror("Writing \"" TICKY " Snap rwxat\"");
		if (writeself2(loadfd, TICKY, "Crackle", "rwx--") < 0)
			perror("Writing \"" TICKY " Crackle rwx--\"");
	} else {
		if (writeself(loadfd, "Pop", "Crackle", "rwxa-") < 0)
			perror("Writing \"Pop Crackle rwxa-\"");
		if (writeself(loadfd, "Pop", "Crackle", "rwx--") < 0)
			perror("Writing \"Pop Crackle rwx--\"");
		if (writeself(loadfd, "Pop", "Snap", "rwxat") < 0)
			perror("Writing \"Pop Snap rwxat\"");
		if (writeself(loadfd, "Pop", "Crackle", "rwx--") < 0)
			perror("Writing \"Pop Crackle rwx--\"");
	}

	printlist(loadfd, "New list", use2);

	if (testfile == NULL)
		exit(0);

	smack_self(self);
	rc = smack_of_file(testfile, it);
	if (rc < 0) {
		printf("Cannot get label for %s\n", testfile);
		exit(1);
	}
	
	if (strcmp(self, it) == 0) {
		printf("Cannot set a rule with %s and %s\n", self, it);
		exit(1);
	}

	if (use2) {
		loadfd = openload2(loadfd);
		if (writeself2(loadfd, self, it, "-----") < 0)
			perror("Writing \"self it -----\"");
	} else {
		loadfd = openload(loadfd);
		if (writeself(loadfd, self, it, "-----") < 0)
			perror("Writing \"self it -----\"");
	}

	printlist(loadfd, "Updated list", use2);

	rc = smack_of_file(testfile, it);
	if (rc >= 0) {
		printf("Unexpected access for %s and %s\n", self, it);
	}

	if (use2) {
		loadfd = openload2(loadfd);
		if (writeself2(loadfd, self, it, "rwxat") < 0)
			perror("Writing \"self it rwxat\"");
	} else {
		loadfd = openload(loadfd);
		if (writeself(loadfd, self, it, "rwxat") < 0)
			perror("Writing \"self it rwxat\"");
	}

	printlist(loadfd, "Updated list", use2);

	rc = smack_of_file(testfile, it);
	if (rc < 0) {
		printf("Unexpected failure for %s and %s\n", self, it);
	}

	/*
	 * Inheritence
	 */
	rc = fork();
	if (rc < 0) {
		perror("Unexpected fork failure.");
		exit(1);
	} 
	if (rc > 0) {
		/*
		 * Parent.
		 */
		if (use2) {
			loadfd = openload2(loadfd);
			if (writeself2(loadfd, "Parent", "Only", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself2(loadfd, "Parent", "Exclusive", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself2(loadfd, "Parent", "NC17", "r-x--") < 0)
				perror("Writing \"self it rwxat\"");
		} else {
			loadfd = openload(loadfd);
			if (writeself(loadfd, "Parent", "Only", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself(loadfd, "Parent", "Exclusive", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself(loadfd, "Parent", "NC17", "r-x--") < 0)
				perror("Writing \"self it rwxat\"");
		}
		rc = wait();
		printlist(loadfd, "Post-child parent list", use2);

	} else {
		/*
		 * Child.
		 */
		printlist(loadfd, "Initial child list", use2);
		if (use2) {
			loadfd = openload2(loadfd);
			if (writeself2(loadfd, "Child", "Only", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself2(loadfd, "Child", "Only", "rwxa-") < 0)
				perror("Writing \"self it rwxat\"");
		} else {
			loadfd = openload(loadfd);
			if (writeself(loadfd, "Child", "Only", "rwxat") < 0)
				perror("Writing \"self it rwxat\"");
			if (writeself(loadfd, "Child", "Only", "rwxa-") < 0)
				perror("Writing \"self it rwxat\"");
		}
		printlist(loadfd, "Later child list", use2);
	}
	exit(0);
}
