/*
 * Copyright (C) 2011 Casey Schaufler <casey@schaufler-ca.com>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2.
 *
 * Author:
 *	Casey Schaufler <casey@schaufler-ca.com>
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/xattr.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LSIZE 23
#define ASIZE 5

#define	L_FILE		"ObjectLabel"
#define	L_LIB		"LibraryLabel"
#define	L_TASK		"TaskLabel"

#define LOADOLD		"/smack/load"
#define LOADSELFOLD	"/smack/load-self"
#define LOAD		"/sys/fs/smackfs/load"
#define LOADSELF	"/sys/fs/smackfs/load-self"

#define MMAPABLE	"/tmp/testmmap"

int writerule(int fd, char *sp, char *op, char *real)
{
	char rule[LSIZE + LSIZE + ASIZE + 3];
	int rc;

	sprintf(rule, "%-23s %-23s %-5s", sp, op, real);
/*
	printf("w\t%s\n", rule);
*/
	rc = lseek(fd, 0, SEEK_SET);
	if (rc < 0)
		perror("lseek in writerule");

	rc = write(fd, rule, LSIZE + LSIZE + ASIZE + 2);
	if (rc < 0)
		perror("write in writerule");

	return rc;
}

int offset;
char buffer[512];

int readself(int fd, char *sp, char *op, char *real)
{
	int red;
	int rc;

	if (offset == 0) {
		bzero(buffer, 512);
		red = read(fd, buffer, 512);
		if (red < 0) {
			perror("reading " LOADSELF);
			return red;
		}
		if (red == 0)
			return 0;
		buffer[red] = '\0';
	}
	if ((rc = sscanf(buffer + offset, "%s %s %s", sp, op, real)) != 3) {
		if (rc >= 0)
			printf("scanned %d\n", rc);
		rc = 0;
	}
	while (buffer[offset] != '\n' && buffer[offset] != '\0')
		offset++;
	if (buffer[offset] == '\n')
		offset++;
	if (buffer[offset] == '\0')
		offset = 0;
	if (offset >= red)
		offset = 0;
	return rc;
}

int openload(int oldfd)
{
	int loadfd;

	if (oldfd >= 0)
		close(oldfd);

	loadfd = open(LOAD, O_RDWR);
	if (loadfd < 0)
		loadfd = open(LOADOLD, O_RDWR);
	if (loadfd < 0) {
		perror("opening " LOAD);
		exit(1);
	}
	return loadfd;
}

int openloadself(int oldfd)
{
	int loadfd;

	if (oldfd >= 0)
		close(oldfd);

	loadfd = open(LOADSELF, O_RDWR);
	if (loadfd < 0)
		loadfd = open(LOADSELFOLD, O_RDWR);
	if (loadfd < 0) {
		perror("opening " LOADSELF);
		exit(1);
	}
	return loadfd;
}

int opencurrent(void)
{
	int fd;

	fd = open("/proc/self/attr/smack.current", O_RDWR);
	if (fd < 0)
		fd = open("/smack/current", O_RDWR);
	if (fd < 0)
		fd = open("/proc/self/attr/current", O_RDWR);
	if (fd < 0) {
		perror("opening /proc/self/attr/current");
		exit(1);
	}
	return fd;
}

void proclabel(int currentfd, char *smack)
{
	int rc;
	char *cp;

	rc = read(currentfd, smack, LSIZE);
	if (rc <= 0) {
		perror("reading /proc/self/attr/current");
		exit(1);
	}
	smack[rc] = 0;
	if ((cp = strchr(smack, '\n')) != NULL)
		*cp = '\0';
}

int smack_set_mmap(char *path, char *msmack)
{
	int rc;

	rc = lsetxattr(path, "security.SMACK64MMAP", msmack, strlen(msmack), 0);
	if (rc <= 0) {
		perror("setting MMAP label");
		exit(1);
	}
	
}

int smack_get_mmap(char *path, char *msmack)
{
	int rc;

	rc = lgetxattr(path, "security.SMACK64MMAP", msmack, LSIZE);
	if (rc > 0)
		msmack[rc] = '\0';
	return rc;
}

int smack_of_file(char *path, char *smack)
{
	int rc;

	rc = lgetxattr(path, "security.SMACK64", smack, LSIZE);
	if (rc > 0)
		smack[rc] = '\0';
	return rc;
}

int mmapfd(char *text)
{
	int fd;
	int rc;
	int i;

	rc = unlink(MMAPABLE);
	if (rc < 0)
		perror("Note on pre-cleaning");

	fd = open(MMAPABLE, O_CREAT | O_RDWR, S_IRWXU | S_IRWXG | S_IRWXO);
	if (fd < 0) {
		perror("creating MMAP file");
		exit(1);
	}
	rc = chmod(MMAPABLE, 0777);
	if (rc < 0)
		perror("setting mode");
	rc = lsetxattr(MMAPABLE, "security.SMACK64", L_FILE, strlen(L_FILE), 0);
	if (rc < 0)
		perror("setting object label");
	rc = lsetxattr(MMAPABLE, "security.SMACK64MMAP", L_LIB, strlen(L_LIB), 0);
	if (rc < 0)
		perror("setting MMAP");

	for (i = 0; i < 1024; i += rc) {
		rc = write(fd, text, strlen(text));
		if (rc <= 0) {
			perror("writing MMAP file");
			exit(1);
		}
	}

	return fd;
}

void printlist(int fd, char *title)
{
	char subject[LSIZE+1];
	char object[LSIZE+1];
	char accesses[ASIZE+1];
	int rc;

	printf("%s:\n", title);
	offset = 0;
	rc = lseek(fd, 0, SEEK_SET);
	if (rc < 0)
		perror("lseek in printlist");

	while (readself(fd, subject, object, accesses) > 0)
		printf("\t%-23s %-23s %-5s\n", subject, object, accesses);
/*
		printf("r\t%-23s %-23s %-5s\n", subject, object, accesses);
*/
}

void tryadding(int rulefd, int mmfd, char *sub, char *obj, char *access)
{
	char *rp;
	char note[300];

	sprintf(note, "New Rule: %s %s %s", sub, obj, access);

	if (writerule(rulefd, sub, obj, access) < 0)
		perror(note);
/*
	printlist(rulefd, note);
*/

	rp = mmap(NULL, 1024, PROT_READ | PROT_WRITE, MAP_SHARED, mmfd, 0);
	if (rp != MAP_FAILED) {
		munmap(rp, 1024);
		printf("%s: Success\n", note);
	}
	else
		perror(note);
}

int
main(int argc, char *argv[])
{
	int rc;
	int loadfd;
	int selffd;
	int currentfd;
	int mmfd;
	char subject[LSIZE+1];
	char object[LSIZE+1];
	char accesses[ASIZE+1];
	char self[LSIZE+1];
	char it[LSIZE+1];
	char *rp;

	loadfd = openload(-1);
	selffd = openloadself(-1);
	currentfd = opencurrent();

	rc = write(currentfd, L_TASK, strlen(L_TASK));
	if (rc < 0)
		perror("write to currentfd");

	mmfd = mmapfd("First Test\n");
	if (mmfd < 0)
		perror("Opening test file");

	printf("Writing global\n");
	if (writerule(loadfd, L_TASK, L_FILE, "rw---") < 0)
		perror("Writing global task file rw---\"");
	if (writerule(loadfd, L_LIB, L_FILE, "rw---") < 0)
		perror("Writing global task file rw---\"");

	printlist(loadfd, "Initial global list");
	printlist(selffd, "Initial local list");

	setreuid(0, 500);

	rp = mmap(NULL, 1024, PROT_READ | PROT_WRITE, MAP_SHARED, mmfd, 0);
	if (rp == MAP_FAILED)
		perror("control mmap failed");
	else {
		printf("mmap with no local rules succeeded\n");
		munmap(rp, 1024);
	}
	
	tryadding(selffd, mmfd, L_TASK, L_LIB, "rw---");
	tryadding(selffd, mmfd, L_TASK, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_TASK, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_FILE, L_TASK, "rw---");
	tryadding(selffd, mmfd, L_FILE, L_LIB, "rw---");
	
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");

	tryadding(selffd, mmfd, L_TASK, L_LIB, "-----");
	tryadding(selffd, mmfd, L_TASK, L_FILE, "-----");
	tryadding(selffd, mmfd, L_LIB, L_TASK, "-----");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");
	tryadding(selffd, mmfd, L_FILE, L_TASK, "-----");
	tryadding(selffd, mmfd, L_FILE, L_LIB, "-----");
	
	tryadding(selffd, mmfd, L_TASK, L_LIB, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	tryadding(selffd, mmfd, L_TASK, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	tryadding(selffd, mmfd, L_LIB, L_TASK, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	tryadding(selffd, mmfd, L_FILE, L_TASK, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	tryadding(selffd, mmfd, L_FILE, L_LIB, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "rw---");
	tryadding(selffd, mmfd, L_LIB, L_FILE, "-----");

	exit(0);
}
